/*/../bin/true; p="/tmp/$0.elf"; v=1; gcc -Wfatal-errors -O42 "$0" -o "$p" && chmod 755 "$p" && ("$p"; v=$?; rm "$p"); exit $v;*/

/* memfill.c - memory allocator behavior tester,
 *             ported from memfill.cpp (please use that instead)
 * 
 * 2018 ed <irc.rizon.net>, MIT licensed
 * https://ocv.me/dev/memfill.c
 * 
 * usage: ./memfill.c
 * 
 * allocating memory:
 *   n  s>  n  s>  n  s>  [...]
 * 
 * freeing memory:
 *   d>  d>  d>  d>  [...]
 * 
 * monitoring memory usage:
 *   pid=''; while true; do [ $pid ] || pid=$(pidof memfill); [ $pid ] || { echo 'x'; sleep 1; continue; }; grep VmRSS /proc/$pid/status || pid='' ; sleep 0.2; done
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define L fprintf(stderr, "\033[35m# %d\033[0m\n", __LINE__);

typedef struct Vector
{
	size_t size;
	size_t ubound;
	char** data;
};


struct Vector* new_vec()
{
	struct Vector* v = malloc(sizeof(struct Vector));
	v->size = 0;
	v->ubound = 4;
	v->data = malloc(v->ubound * sizeof(char*));
	if (v->data)
		return v;
	
	return 0;
}


int vec_push(struct Vector* v, char* item)
{
	if (v->size >= v->ubound)
	{
		v->ubound *= 2;
		v->data = realloc(v->data, v->ubound * sizeof(char*));
		if (!v->data)
			return -1;
	}
	v->data[v->size++] = item;
	return 0;
}


int vec_pop(struct Vector* v)
{
	if (v->size == 0)
		return -1;
	
	--v->size;
	return 0;
}


int vec_remove(struct Vector* v, size_t idx)
{
	if (idx >= v->size)
		return -1;
	
	for (++idx; idx < v->size; idx++)
		v->data[idx-1] = v->data[idx];
	
	--v->size;
}


size_t getint(char* cmd, int min, int max)
{
	if (strlen(cmd) < 2)
	{
		printf("command needs a numeric argument\n");
		return -1;
	}
	
	if (cmd[1] == '<')
		return min;
	
	if (cmd[1] == '>')
		return max;
	
	char* it = cmd;
	while (it[0] != '\0' && it[0] >= '0' && it[0] <= '9')
		++it;
	
	if (it[0] != '\0')
	{
		printf("bad argument; must be a number OR '<' for min (%d) OR '>' for max (%d): %s\n",
			min, max, cmd+1);
		
		return -1;
	}
	
	size_t ret = atol(cmd+1);
	if (ret < min)
	{
		printf("argument below minimum: %lu < %d\n", ret, min);
		return -1;
	}
	if (ret > max)
	{
		printf("argument above maximum: %lu > %d\n", ret, max);
		return -1;
	}
	return ret;
}


int main(int argc, char** argv)
{
	size_t blocksize = 1024;
	blocksize *= 1024;
	blocksize *= 16;
	printf("using blocksize %lu bytes = %lu kB = %lu MB\n",
		blocksize, blocksize/1024, blocksize/(1024*1024));
	
	char* cmd = 0;
	size_t cmdlen = -1;
	
	unsigned long malloced = 0;
	struct Vector* blocks = new_vec();
	
	for(;;)
	{
		printf("\n[n]ew [r]ead [p]oke [s]et [u]nset [d]elete [q]uit\ncmd> ");
		fflush(stdout);
		if (getline(&cmd, &cmdlen, stdin) == -1)
			break;
		
		if (cmdlen == 0)
			continue;
		
		if (cmd[0] == 'q')
			break;
		
		if (cmd[0] == 'n')
		{
			char* block = (char*)malloc(blocksize);
			vec_push(blocks, block);
			malloced += blocksize;
			printf("created block %lu, %luMB malloced\n",
				blocks->size-1, malloced/(1024*1024));
			
			continue;
		}

		int idx = getint(cmd, 0, blocks->size - 1);
		if (idx < 0)
			continue;

		if (cmd[0] == 'r')
		{
			printf("reading values from block %d\n", idx);
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
			{
				size_t v = *((int*)&(blocks->data[idx][ofs]));
				v += 1;
			}
			continue;
		}

		if (cmd[0] == 'p')
		{
			printf("setting 256 values (1MB) in block %d\n", idx);
			for (size_t ofs = 0; ofs < 0x100000; ofs += 0x1000)
				blocks->data[idx][ofs] = ofs/0x1000;

			continue;
		}

		if (cmd[0] == 's')
		{
			printf("setting values in block %d\n", idx);
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
				*((int*)&(blocks->data[idx][ofs])) = ofs;
			
			//for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
			//	std::cout << std::hex << *((int*)&(blocks->data[idx][ofs])) << " ";

			continue;
		}

		if (cmd[0] == 'u')
		{
			printf("blanking values in block %d\n", idx);
			size_t nil = 0;
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
				*((int*)&(blocks->data[idx][ofs])) = nil;

			continue;
		}

		if (cmd[0] == 'd')
		{
			malloced -= blocksize;
			printf("erasing block at index %d, %luMB malloced\n",
				idx, malloced/(1024*1024));
			
			char* block = blocks->data[idx];
			free(block);
			vec_remove(blocks, idx);

			continue;
		}
	}
	
	for (size_t ofs = blocks->size; ofs > 0; --ofs)
	{
		free(blocks->data[ofs-1]);
		vec_pop(blocks);
	}
	free(blocks->data);
	free(blocks);
	free(cmd);
}
